;; Generated package description from denote-review.el  -*- mode: lisp-data; no-byte-compile: t -*-
(define-package "denote-review" "1.0.4" "implements review process for denote notes" '((emacs "28.1") (denote "4.1.3")) :commit "9a4cf8a422e1bc915b2624a34ef39563709915fd" :authors '(("Matto Fransen" . "matto@matto.nl")) :maintainer '("Matto Fransen" . "matto@matto.nl") :keywords '("files") :url "https://codeberg.org/mattof/denote-review")
